@@ -1,3 +1,4 @@
+use crate::error::ResamplerConstructionError;
 use crate::sinc::make_sincs;
 use crate::windows::WindowFunction;
 use num_complex::Complex;
@@ -6,7 +7,7 @@ use num_traits::Zero;
 use std::sync::Arc;
 
 use crate::error::{ResampleError, ResampleResult};
-use crate::{Resampler, Sample};
+use crate::{update_mask_from_buffers, validate_buffers, Resampler, Sample};
 use realfft::{ComplexToReal, RealFftPlanner, RealToComplex};
 
 /// A helper for resampling a single chunk of data.
@@ -37,6 +38,7 @@ pub struct FftFixedIn<T> {
     fft_size_out: usize,
     overlaps: Vec<Vec<T>>,
     input_buffers: Vec<Vec<T>>,
+    channel_mask: Vec<bool>,
     saved_frames: usize,
     resampler: FftResampler<T>,
 }
@@ -54,6 +56,7 @@ pub struct FftFixedOut<T> {
     fft_size_out: usize,
     overlaps: Vec<Vec<T>>,
     output_buffers: Vec<Vec<T>>,
+    channel_mask: Vec<bool>,
     saved_frames: usize,
     frames_needed: usize,
     resampler: FftResampler<T>,
@@ -70,10 +73,18 @@ pub struct FftFixedInOut<T> {
     chunk_size_in: usize,
     chunk_size_out: usize,
     fft_size_in: usize,
+    channel_mask: Vec<bool>,
     overlaps: Vec<Vec<T>>,
     resampler: FftResampler<T>,
 }
 
+fn validate_sample_rates(input: usize, output: usize) -> Result<(), ResamplerConstructionError> {
+    if input == 0 || output == 0 {
+        return Err(ResamplerConstructionError::InvalidSampleRate { input, output });
+    }
+    Ok(())
+}
+
 impl<T> FftResampler<T>
 where
     T: Sample,
@@ -82,7 +93,7 @@ where
     pub fn new(fft_size_in: usize, fft_size_out: usize) -> Self {
         // calculate antialiasing cutoff
         let cutoff = if fft_size_in > fft_size_out {
-            0.4f32.powf(16.0 / fft_size_in as f32) * fft_size_out as f32 / fft_size_in as f32
+            0.4f32.powf(16.0 / fft_size_out as f32) * fft_size_out as f32 / fft_size_in as f32
         } else {
             0.4f32.powf(16.0 / fft_size_in as f32)
         };
@@ -141,25 +152,24 @@ where
             .process_with_scratch(&mut self.input_buf, &mut self.input_f, &mut self.scratch_fw)
             .unwrap();
 
-        // multiply with filter FT
-        self.input_f
-            .iter_mut()
-            .take(self.fft_size_in + 1)
-            .zip(self.filter_f.iter())
-            .for_each(|(spec, filt)| *spec *= filt);
         let new_len = if self.fft_size_in < self.fft_size_out {
             self.fft_size_in + 1
         } else {
             self.fft_size_out
         };
 
+        // multiply with filter FT
+        self.input_f
+            .iter_mut()
+            .take(new_len)
+            .zip(self.filter_f.iter())
+            .for_each(|(spec, filt)| *spec *= filt);
+
         // copy to modified spectrum
         self.output_f[0..new_len].copy_from_slice(&self.input_f[0..new_len]);
         for val in self.output_f[new_len..].iter_mut() {
             *val = Complex::zero();
         }
-        //self.output_f[self.fft_size_out] = self.input_f[self.fft_size_in];
-
         // IFFT result, store result and overlap
         self.ifft
             .process_with_scratch(
@@ -182,35 +192,45 @@ where
     /// Create a new FftFixedInOut
     ///
     /// Parameters are:
-    /// - `fs_in`: Input sample rate.
-    /// - `fs_out`: Output sample rate.
+    /// - `sample_rate_input`: Input sample rate, must be > 0.
+    /// - `sample_rate_output`: Output sample rate, must be > 0.
     /// - `chunk_size_in`: desired length of input data in frames, actual value may be different.
     /// - `nbr_channels`: number of channels in input/output.
-    pub fn new(fs_in: usize, fs_out: usize, chunk_size_in: usize, nbr_channels: usize) -> Self {
+    pub fn new(
+        sample_rate_input: usize,
+        sample_rate_output: usize,
+        chunk_size_in: usize,
+        nbr_channels: usize,
+    ) -> Result<Self, ResamplerConstructionError> {
+        validate_sample_rates(sample_rate_input, sample_rate_output)?;
+
         debug!(
-            "Create new FftFixedInOut, fs_in: {}, fs_out: {} chunk_size_in: {}, channels: {}",
-            fs_in, fs_out, chunk_size_in, nbr_channels
+            "Create new FftFixedInOut, sample_rate_input: {}, sample_rate_output: {} chunk_size_in: {}, channels: {}",
+            sample_rate_input, sample_rate_output, chunk_size_in, nbr_channels
         );
 
-        let gcd = integer::gcd(fs_in, fs_out);
-        let min_chunk_out = fs_out / gcd;
+        let gcd = integer::gcd(sample_rate_input, sample_rate_output);
+        let min_chunk_out = sample_rate_output / gcd;
         let wanted = chunk_size_in;
         let fft_chunks = (wanted as f32 / min_chunk_out as f32).ceil() as usize;
-        let fft_size_out = fft_chunks * fs_out / gcd;
-        let fft_size_in = fft_chunks * fs_in / gcd;
+        let fft_size_out = fft_chunks * sample_rate_output / gcd;
+        let fft_size_in = fft_chunks * sample_rate_input / gcd;
 
         let resampler = FftResampler::<T>::new(fft_size_in, fft_size_out);
 
         let overlaps: Vec<Vec<T>> = vec![vec![T::zero(); fft_size_out]; nbr_channels];
 
-        FftFixedInOut {
+        let channel_mask = vec![true; nbr_channels];
+
+        Ok(FftFixedInOut {
             nbr_channels,
             chunk_size_in: fft_size_in,
             chunk_size_out: fft_size_out,
             fft_size_in,
             overlaps,
             resampler,
-        }
+            channel_mask,
+        })
     }
 }
 
@@ -218,62 +238,75 @@ impl<T> Resampler<T> for FftFixedInOut<T>
 where
     T: Sample,
 {
-    /// Query for the number of frames needed for the next call to "process".
-    fn nbr_frames_needed(&self) -> usize {
+    fn process_into_buffer<V: AsRef<[T]>>(
+        &mut self,
+        wave_in: &[V],
+        wave_out: &mut [Vec<T>],
+        active_channels_mask: Option<&[bool]>,
+    ) -> ResampleResult<()> {
+        if let Some(mask) = active_channels_mask {
+            self.channel_mask.copy_from_slice(mask);
+        } else {
+            update_mask_from_buffers(wave_in, &mut self.channel_mask);
+        };
+
+        validate_buffers(
+            wave_in,
+            wave_out,
+            &self.channel_mask,
+            self.nbr_channels,
+            self.chunk_size_in,
+        )?;
+
+        for (chan, active) in self.channel_mask.iter().enumerate() {
+            if *active {
+                if self.chunk_size_out > wave_out[chan].capacity() {
+                    trace!(
+                        "Allocating more space for channel {}, old capacity: {}, new: {}",
+                        chan,
+                        wave_out[chan].capacity(),
+                        self.chunk_size_out
+                    );
+                }
+                wave_out[chan].resize(self.chunk_size_out, T::zero());
+                self.resampler.resample_unit(
+                    wave_in[chan].as_ref(),
+                    &mut wave_out[chan],
+                    &mut self.overlaps[chan],
+                )
+            }
+        }
+        Ok(())
+    }
+
+    fn input_frames_max(&self) -> usize {
         self.fft_size_in
     }
 
-    /// Resample a chunk of audio. The input and output lengths are fixed.
-    /// If the waveform for a channel is empty, this channel will be ignored and produce a
-    /// corresponding empty output waveform.
-    /// # Errors
-    ///
-    /// The function returns an error if the size of the input data is not equal
-    /// to the number of channels and input size defined when creating the instance.
-    fn process<V: AsRef<[T]>>(&mut self, wave_in: &[V]) -> ResampleResult<Vec<Vec<T>>> {
-        if wave_in.len() != self.nbr_channels {
-            return Err(ResampleError::WrongNumberOfChannels {
-                expected: self.nbr_channels,
-                actual: wave_in.len(),
-            });
-        }
-        let mut used_channels = Vec::new();
-        for (chan, wave) in wave_in.iter().enumerate() {
-            let wave = wave.as_ref();
-            if !wave.is_empty() {
-                used_channels.push(chan);
-                if wave.len() != self.chunk_size_in {
-                    return Err(ResampleError::WrongNumberOfFrames {
-                        channel: chan,
-                        expected: self.chunk_size_in,
-                        actual: wave.len(),
-                    });
-                }
-            }
-        }
-        let mut wave_out = vec![Vec::new(); self.nbr_channels];
-        for chan in used_channels.iter() {
-            wave_out[*chan] = vec![T::zero(); self.chunk_size_out];
-        }
+    fn input_frames_next(&self) -> usize {
+        self.fft_size_in
+    }
 
-        for n in used_channels.iter() {
-            self.resampler.resample_unit(
-                wave_in[*n].as_ref(),
-                &mut wave_out[*n],
-                &mut self.overlaps[*n],
-            )
-        }
-        Ok(wave_out)
+    fn nbr_channels(&self) -> usize {
+        self.nbr_channels
+    }
+
+    fn output_frames_max(&self) -> usize {
+        self.chunk_size_out
+    }
+
+    fn output_frames_next(&self) -> usize {
+        self.output_frames_max()
     }
 
     /// Update the resample ratio. This is not supported by this resampler and
-    /// always returns an error.
+    /// always returns an [ResampleError::SyncNotAdjustable].
     fn set_resample_ratio(&mut self, _new_ratio: f64) -> ResampleResult<()> {
         Err(ResampleError::SyncNotAdjustable)
     }
 
     /// Update the resample ratio relative to the original one. This is not
-    /// supported by this resampler and always returns an error.
+    /// supported by this resampler and always returns an [ResampleError::SyncNotAdjustable].
     fn set_resample_ratio_relative(&mut self, _rel_ratio: f64) -> ResampleResult<()> {
         Err(ResampleError::SyncNotAdjustable)
     }
@@ -286,41 +319,45 @@ where
     /// Create a new FftFixedOut
     ///
     /// Parameters are:
-    /// - `fs_in`: Input sample rate.
-    /// - `fs_out`: Output sample rate.
+    /// - `sample_rate_input`: Input sample rate, must be > 0.
+    /// - `sample_rate_output`: Output sample rate, must be > 0.
     /// - `chunk_size_out`: length of output data in frames.
     /// - `sub_chunks`: desired number of subchunks for processing, actual number may be different.
     /// - `nbr_channels`: number of channels in input/output.
     pub fn new(
-        fs_in: usize,
-        fs_out: usize,
+        sample_rate_input: usize,
+        sample_rate_output: usize,
         chunk_size_out: usize,
         sub_chunks: usize,
         nbr_channels: usize,
-    ) -> Self {
-        let gcd = integer::gcd(fs_in, fs_out);
-        let min_chunk_out = fs_out / gcd;
+    ) -> Result<Self, ResamplerConstructionError> {
+        validate_sample_rates(sample_rate_input, sample_rate_output)?;
+
+        let gcd = integer::gcd(sample_rate_input, sample_rate_output);
+        let min_chunk_out = sample_rate_output / gcd;
         let wanted_subsize = chunk_size_out / sub_chunks;
         let fft_chunks = (wanted_subsize as f32 / min_chunk_out as f32).ceil() as usize;
-        let fft_size_out = fft_chunks * fs_out / gcd;
-        let fft_size_in = fft_chunks * fs_in / gcd;
+        let fft_size_out = fft_chunks * sample_rate_output / gcd;
+        let fft_size_in = fft_chunks * sample_rate_input / gcd;
 
         let resampler = FftResampler::<T>::new(fft_size_in, fft_size_out);
 
         debug!(
-            "Create new FftFixedOut, fs_in: {}, fs_out: {} chunk_size_in: {}, channels: {}, fft_size_in: {}, fft_size_out: {}",
-            fs_in, fs_out, chunk_size_out, nbr_channels, fft_size_in, fft_size_out
+            "Create new FftFixedOut, sample_rate_input: {}, sample_rate_output: {} chunk_size_in: {}, channels: {}, fft_size_in: {}, fft_size_out: {}",
+            sample_rate_input, sample_rate_output, chunk_size_out, nbr_channels, fft_size_in, fft_size_out
         );
 
         let overlaps: Vec<Vec<T>> = vec![vec![T::zero(); fft_size_out]; nbr_channels];
         let output_buffers: Vec<Vec<T>> =
             vec![vec![T::zero(); chunk_size_out + fft_size_out]; nbr_channels];
 
+        let channel_mask = vec![true; nbr_channels];
+
         let saved_frames = 0;
         let chunks_needed = (chunk_size_out as f32 / fft_size_out as f32).ceil() as usize;
         let frames_needed = chunks_needed * fft_size_in;
 
-        FftFixedOut {
+        Ok(FftFixedOut {
             nbr_channels,
             chunk_size_out,
             fft_size_in,
@@ -330,7 +367,8 @@ where
             saved_frames,
             frames_needed,
             resampler,
-        }
+            channel_mask,
+        })
     }
 }
 
@@ -338,71 +376,63 @@ impl<T> Resampler<T> for FftFixedOut<T>
 where
     T: Sample,
 {
-    /// Query for the number of frames needed for the next call to "process".
-    fn nbr_frames_needed(&self) -> usize {
-        self.frames_needed
-    }
+    fn process_into_buffer<V: AsRef<[T]>>(
+        &mut self,
+        wave_in: &[V],
+        wave_out: &mut [Vec<T>],
+        active_channels_mask: Option<&[bool]>,
+    ) -> ResampleResult<()> {
+        if let Some(mask) = active_channels_mask {
+            self.channel_mask.copy_from_slice(mask);
+        } else {
+            update_mask_from_buffers(wave_in, &mut self.channel_mask);
+        };
 
-    /// Resample a chunk of audio. The required input length is provided by
-    /// the "nbr_frames_needed" function, and the output length is fixed.
-    /// If the waveform for a channel is empty, this channel will be ignored and produce a
-    /// corresponding empty output waveform.
-    /// # Errors
-    ///
-    /// The function returns an error if the length of the input data is not
-    /// equal to the number of channels defined when creating the instance,
-    /// and the number of audio frames given by "nbr_frames_needed".
-    fn process<V: AsRef<[T]>>(&mut self, wave_in: &[V]) -> ResampleResult<Vec<Vec<T>>> {
-        if wave_in.len() != self.nbr_channels {
-            return Err(ResampleError::WrongNumberOfChannels {
-                expected: self.nbr_channels,
-                actual: wave_in.len(),
-            });
-        }
-        let mut used_channels = Vec::new();
-        for (chan, wave) in wave_in.iter().enumerate() {
-            let wave = wave.as_ref();
-            if !wave.is_empty() {
-                used_channels.push(chan);
-                if wave.len() != self.frames_needed {
-                    return Err(ResampleError::WrongNumberOfFrames {
-                        channel: chan,
-                        expected: self.frames_needed,
-                        actual: wave.len(),
-                    });
+        validate_buffers(
+            wave_in,
+            wave_out,
+            &self.channel_mask,
+            self.nbr_channels,
+            self.frames_needed,
+        )?;
+
+        for (chan, active) in self.channel_mask.iter().enumerate() {
+            if *active {
+                if self.chunk_size_out > wave_out[chan].capacity() {
+                    trace!(
+                        "Allocating more space for channel {}, old capacity: {}, new: {}",
+                        chan,
+                        wave_out[chan].capacity(),
+                        self.chunk_size_out
+                    );
+                }
+                wave_out[chan].resize(self.chunk_size_out, T::zero());
+                for (in_chunk, out_chunk) in wave_in[chan].as_ref().chunks(self.fft_size_in).zip(
+                    self.output_buffers[chan][self.saved_frames..].chunks_mut(self.fft_size_out),
+                ) {
+                    self.resampler
+                        .resample_unit(in_chunk, out_chunk, &mut self.overlaps[chan]);
                 }
-            }
-        }
-
-        let mut wave_out = vec![Vec::new(); self.nbr_channels];
-        for chan in used_channels.iter() {
-            wave_out[*chan] = self.output_buffers[*chan].clone();
-        }
-
-        for n in used_channels.iter() {
-            for (in_chunk, out_chunk) in wave_in[*n]
-                .as_ref()
-                .chunks(self.fft_size_in)
-                .zip(wave_out[*n][self.saved_frames..].chunks_mut(self.fft_size_out))
-            {
-                self.resampler
-                    .resample_unit(in_chunk, out_chunk, &mut self.overlaps[*n]);
             }
         }
         let processed_frames =
             self.saved_frames + self.fft_size_out * (self.frames_needed / self.fft_size_in);
 
-        // save extra frames for next round
-        self.saved_frames = processed_frames - self.chunk_size_out;
-        if processed_frames > self.chunk_size_out {
-            for n in used_channels.iter() {
-                self.output_buffers[*n][0..self.saved_frames].copy_from_slice(
-                    &wave_out[*n][self.chunk_size_out..(self.chunk_size_out + self.saved_frames)],
-                );
+        // copy to output, and save extra frames for next round
+        if processed_frames >= self.chunk_size_out {
+            self.saved_frames = processed_frames - self.chunk_size_out;
+            for (chan, active) in self.channel_mask.iter().enumerate() {
+                if *active {
+                    wave_out[chan][..]
+                        .copy_from_slice(&self.output_buffers[chan][0..self.chunk_size_out]);
+                    self.output_buffers[chan].copy_within(
+                        self.chunk_size_out..(self.chunk_size_out + self.saved_frames),
+                        0,
+                    );
+                }
             }
-        }
-        for n in used_channels.iter() {
-            wave_out[*n].truncate(self.chunk_size_out);
+        } else {
+            self.saved_frames = processed_frames;
         }
         //calculate number of needed frames from next round
         let frames_needed_out = if self.chunk_size_out > self.saved_frames {
@@ -412,17 +442,37 @@ where
         };
         let chunks_needed = (frames_needed_out as f32 / self.fft_size_out as f32).ceil() as usize;
         self.frames_needed = chunks_needed * self.fft_size_in;
-        Ok(wave_out)
+        Ok(())
+    }
+
+    fn input_frames_max(&self) -> usize {
+        (self.chunk_size_out as f32 / self.fft_size_out as f32).ceil() as usize * self.fft_size_in
+    }
+
+    fn input_frames_next(&self) -> usize {
+        self.frames_needed
+    }
+
+    fn nbr_channels(&self) -> usize {
+        self.nbr_channels
+    }
+
+    fn output_frames_max(&self) -> usize {
+        self.chunk_size_out
+    }
+
+    fn output_frames_next(&self) -> usize {
+        self.output_frames_max()
     }
 
     /// Update the resample ratio. This is not supported by this resampler and
-    /// always returns an error.
+    /// always returns [ResampleError::SyncNotAdjustable].
     fn set_resample_ratio(&mut self, _new_ratio: f64) -> ResampleResult<()> {
         Err(ResampleError::SyncNotAdjustable)
     }
 
     /// Update the resample ratio relative to the original one. This is not
-    /// supported by this resampler and always returns an error.
+    /// supported by this resampler and always returns [ResampleError::SyncNotAdjustable].
     fn set_resample_ratio_relative(&mut self, _rel_ratio: f64) -> ResampleResult<()> {
         Err(ResampleError::SyncNotAdjustable)
     }
@@ -432,41 +482,45 @@ impl<T> FftFixedIn<T>
 where
     T: Sample,
 {
-    /// Create a new FftFixedOut
+    /// Create a new FftFixedIn
     ///
     /// Parameters are:
-    /// - `fs_in`: Input sample rate.
-    /// - `fs_out`: Output sample rate.
-    /// - `chunk_size_out`: length of output data in frames.
+    /// - `sample_rate_input`: Input sample rate, must be > 0.
+    /// - `sample_rate_output`: Output sample rate, must be > 0.
+    /// - `chunk_size_in`: length of input data in frames.
     /// - `sub_chunks`: desired number of subchunks for processing, actual number used may be different.
     /// - `nbr_channels`: number of channels in input/output.
     pub fn new(
-        fs_in: usize,
-        fs_out: usize,
+        sample_rate_input: usize,
+        sample_rate_output: usize,
         chunk_size_in: usize,
         sub_chunks: usize,
         nbr_channels: usize,
-    ) -> Self {
-        let gcd = integer::gcd(fs_in, fs_out);
-        let min_chunk_in = fs_in / gcd;
+    ) -> Result<Self, ResamplerConstructionError> {
+        validate_sample_rates(sample_rate_input, sample_rate_output)?;
+
+        let gcd = integer::gcd(sample_rate_input, sample_rate_output);
+        let min_chunk_in = sample_rate_input / gcd;
         let wanted_subsize = chunk_size_in / sub_chunks;
         let fft_chunks = (wanted_subsize as f32 / min_chunk_in as f32).ceil() as usize;
-        let fft_size_out = fft_chunks * fs_out / gcd;
-        let fft_size_in = fft_chunks * fs_in / gcd;
+        let fft_size_out = fft_chunks * sample_rate_output / gcd;
+        let fft_size_in = fft_chunks * sample_rate_input / gcd;
 
         let resampler = FftResampler::<T>::new(fft_size_in, fft_size_out);
         debug!(
-            "Create new FftFixedOut, fs_in: {}, fs_out: {} chunk_size_in: {}, channels: {}, fft_size_in: {}, fft_size_out: {}",
-            fs_in, fs_out, chunk_size_in, nbr_channels, fft_size_in, fft_size_out
+            "Create new FftFixedOut, sample_rate_input: {}, sample_rate_output: {} chunk_size_in: {}, channels: {}, fft_size_in: {}, fft_size_out: {}",
+            sample_rate_input, sample_rate_output, chunk_size_in, nbr_channels, fft_size_in, fft_size_out
         );
 
         let overlaps: Vec<Vec<T>> = vec![vec![T::zero(); fft_size_out]; nbr_channels];
         let input_buffers: Vec<Vec<T>> =
-            vec![vec![T::zero(); chunk_size_in + fft_size_out]; nbr_channels];
+            vec![vec![T::zero(); chunk_size_in + fft_size_in]; nbr_channels];
+
+        let channel_mask = vec![true; nbr_channels];
 
         let saved_frames = 0;
 
-        FftFixedIn {
+        Ok(FftFixedIn {
             nbr_channels,
             chunk_size_in,
             fft_size_in,
@@ -475,7 +529,8 @@ where
             input_buffers,
             saved_frames,
             resampler,
-        }
+            channel_mask,
+        })
     }
 }
 
@@ -483,83 +538,63 @@ impl<T> Resampler<T> for FftFixedIn<T>
 where
     T: Sample,
 {
-    /// Query for the number of frames needed for the next call to "process".
-    fn nbr_frames_needed(&self) -> usize {
-        self.chunk_size_in
-    }
+    fn process_into_buffer<V: AsRef<[T]>>(
+        &mut self,
+        wave_in: &[V],
+        wave_out: &mut [Vec<T>],
+        active_channels_mask: Option<&[bool]>,
+    ) -> ResampleResult<()> {
+        if let Some(mask) = active_channels_mask {
+            self.channel_mask.copy_from_slice(mask);
+        } else {
+            update_mask_from_buffers(wave_in, &mut self.channel_mask);
+        };
 
-    /// Resample a chunk of audio. The required input length is provided by
-    /// the "nbr_frames_needed" function, and the output length is fixed.
-    /// If the waveform for a channel is empty, this channel will be ignored and produce a
-    /// corresponding empty output waveform.
-    /// # Errors
-    ///
-    /// The function returns an error if the length of the input data is not
-    /// equal to the number of channels defined when creating the instance,
-    /// and the number of audio frames given by "nbr_frames_needed".
-    fn process<V: AsRef<[T]>>(&mut self, wave_in: &[V]) -> ResampleResult<Vec<Vec<T>>> {
-        if wave_in.len() != self.nbr_channels {
-            return Err(ResampleError::WrongNumberOfChannels {
-                expected: self.nbr_channels,
-                actual: wave_in.len(),
-            });
-        }
-        let mut used_channels = Vec::new();
-        for (chan, wave) in wave_in.iter().enumerate() {
-            let wave = wave.as_ref();
-            if !wave.is_empty() {
-                used_channels.push(chan);
-                if wave.len() != self.chunk_size_in {
-                    return Err(ResampleError::WrongNumberOfFrames {
-                        channel: chan,
-                        expected: self.chunk_size_in,
-                        actual: wave.len(),
-                    });
-                }
-            }
-        }
-
-        let mut input_temp = vec![Vec::new(); self.nbr_channels];
-        for chan in used_channels.iter() {
-            input_temp[*chan] = vec![T::zero(); self.saved_frames + self.chunk_size_in];
-        }
+        validate_buffers(
+            wave_in,
+            wave_out,
+            &self.channel_mask,
+            self.nbr_channels,
+            self.chunk_size_in,
+        )?;
 
         // copy new samples to input buffer
-        for n in used_channels.iter() {
-            for (input, buffer) in self.input_buffers[*n]
-                .iter()
-                .take(self.saved_frames)
-                .zip(input_temp[*n].iter_mut())
-            {
-                *buffer = *input;
-            }
-        }
-        for n in used_channels.iter() {
-            for (input, buffer) in wave_in[*n].as_ref().iter().zip(
-                input_temp[*n]
-                    .iter_mut()
-                    .skip(self.saved_frames)
-                    .take(self.chunk_size_in),
-            ) {
-                *buffer = *input;
+        for (chan, active) in self.channel_mask.iter().enumerate() {
+            if *active {
+                for (input, buffer) in wave_in[chan].as_ref().iter().zip(
+                    self.input_buffers[chan]
+                        .iter_mut()
+                        .skip(self.saved_frames)
+                        .take(self.chunk_size_in),
+                ) {
+                    *buffer = *input;
+                }
             }
         }
         self.saved_frames += self.chunk_size_in;
 
         let nbr_chunks_ready =
             (self.saved_frames as f32 / self.fft_size_in as f32).floor() as usize;
-        let mut wave_out = vec![Vec::new(); self.nbr_channels];
-        for chan in used_channels.iter() {
-            wave_out[*chan] = vec![T::zero(); nbr_chunks_ready * self.fft_size_out];
-        }
-        for n in used_channels.iter() {
-            for (in_chunk, out_chunk) in input_temp[*n]
-                .chunks(self.fft_size_in)
-                .take(nbr_chunks_ready)
-                .zip(wave_out[*n].chunks_mut(self.fft_size_out))
-            {
-                self.resampler
-                    .resample_unit(in_chunk, out_chunk, &mut self.overlaps[*n]);
+        let needed_len = nbr_chunks_ready * self.fft_size_out;
+        for (chan, active) in self.channel_mask.iter().enumerate() {
+            if *active {
+                if needed_len > wave_out[chan].capacity() {
+                    trace!(
+                        "Allocating more space for channel {}, old capacity: {}, new: {}",
+                        chan,
+                        wave_out[chan].capacity(),
+                        needed_len
+                    );
+                }
+                wave_out[chan].resize(needed_len, T::zero());
+                for (in_chunk, out_chunk) in self.input_buffers[chan]
+                    .chunks(self.fft_size_in)
+                    .take(nbr_chunks_ready)
+                    .zip(wave_out[chan].chunks_mut(self.fft_size_out))
+                {
+                    self.resampler
+                        .resample_unit(in_chunk, out_chunk, &mut self.overlaps[chan]);
+                }
             }
         }
 
@@ -568,29 +603,46 @@ where
         let extra = self.saved_frames - frames_in_used;
 
         if self.saved_frames > frames_in_used {
-            for n in used_channels.iter() {
-                for (input, buffer) in input_temp[*n]
-                    .iter()
-                    .skip(frames_in_used)
-                    .take(extra)
-                    .zip(self.input_buffers[*n].iter_mut())
-                {
-                    *buffer = *input;
+            for (chan, active) in self.channel_mask.iter().enumerate() {
+                if *active {
+                    self.input_buffers[chan].copy_within(frames_in_used..self.saved_frames, 0);
                 }
             }
         }
         self.saved_frames = extra;
-        Ok(wave_out)
+        Ok(())
+    }
+
+    fn input_frames_max(&self) -> usize {
+        self.chunk_size_in
+    }
+
+    fn input_frames_next(&self) -> usize {
+        self.chunk_size_in
+    }
+
+    fn nbr_channels(&self) -> usize {
+        self.nbr_channels
+    }
+
+    fn output_frames_max(&self) -> usize {
+        self.chunk_size_in * (self.fft_size_out / self.fft_size_in + 1)
+    }
+
+    fn output_frames_next(&self) -> usize {
+        (((self.saved_frames + self.chunk_size_in) as f32) / self.fft_size_in as f32).floor()
+            as usize
+            * self.fft_size_out
     }
 
     /// Update the resample ratio. This is not supported by this resampler and
-    /// always returns an error.
+    /// always returns [ResampleError::SyncNotAdjustable].
     fn set_resample_ratio(&mut self, _new_ratio: f64) -> ResampleResult<()> {
         Err(ResampleError::SyncNotAdjustable)
     }
 
     /// Update the resample ratio relative to the original one. This is not
-    /// supported by this resampler and always returns an error.
+    /// supported by this resampler and always returns [ResampleError::SyncNotAdjustable].
     fn set_resample_ratio_relative(&mut self, _rel_ratio: f64) -> ResampleResult<()> {
         Err(ResampleError::SyncNotAdjustable)
     }
@@ -625,10 +677,10 @@ mod tests {
     #[test]
     fn make_resampler_fio() {
         // asking for 1024 give the nearest which is 1029 -> 1120
-        let mut resampler = FftFixedInOut::<f64>::new(44100, 48000, 1024, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedInOut::<f64>::new(44100, 48000, 1024, 2).unwrap();
+        let frames = resampler.input_frames_next();
         let waves = vec![vec![0.0f64; frames]; 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 1120);
     }
@@ -636,10 +688,10 @@ mod tests {
     #[test]
     fn make_resampler_fio_skipped() {
         // asking for 1024 give the nearest which is 1029 -> 1120
-        let mut resampler = FftFixedInOut::<f64>::new(44100, 48000, 1024, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedInOut::<f64>::new(44100, 48000, 1024, 2).unwrap();
+        let frames = resampler.input_frames_next();
         let waves = vec![vec![0.0f64; frames], Vec::new()];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 1120);
         assert!(out[1].is_empty());
@@ -647,22 +699,22 @@ mod tests {
 
     #[test]
     fn make_resampler_fo() {
-        let mut resampler = FftFixedOut::<f64>::new(44100, 192000, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedOut::<f64>::new(44100, 192000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 294);
         let waves = vec![vec![0.0f64; frames]; 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 1024);
     }
 
     #[test]
     fn make_resampler_fo_skipped() {
-        let mut resampler = FftFixedOut::<f64>::new(44100, 192000, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedOut::<f64>::new(44100, 192000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 294);
         let waves = vec![vec![0.0f64; frames], Vec::new()];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 1024);
         assert!(out[1].is_empty());
@@ -670,11 +722,11 @@ mod tests {
 
     #[test]
     fn make_resampler_fo_empty() {
-        let mut resampler = FftFixedOut::<f64>::new(44100, 192000, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedOut::<f64>::new(44100, 192000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 294);
         let waves = vec![Vec::new(); 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert!(out[0].is_empty());
         assert!(out[1].is_empty());
@@ -682,33 +734,48 @@ mod tests {
 
     #[test]
     fn make_resampler_fi() {
-        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 1024);
         let waves = vec![vec![0.0f64; frames]; 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
+        assert_eq!(out.len(), 2);
+        assert_eq!(out[0].len(), 640);
+    }
+
+    #[test]
+    fn make_resampler_fi_noalloc() {
+        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
+        assert_eq!(frames, 1024);
+        let waves = vec![vec![0.0f64; frames]; 2];
+        let mut out = vec![vec![0.0f64; 2 * frames]; 2];
+        let mask = vec![true; 2];
+        resampler
+            .process_into_buffer(&waves, &mut out, Some(&mask))
+            .unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 640);
     }
 
     #[test]
     fn make_resampler_fi_downsample() {
-        let mut resampler = FftFixedIn::<f64>::new(48000, 16000, 1200, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedIn::<f64>::new(48000, 16000, 1200, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 1200);
         let waves = vec![vec![0.0f64; frames]; 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 400);
     }
 
     #[test]
     fn make_resampler_fi_skipped() {
-        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 1024);
         let waves = vec![vec![0.0f64; frames], Vec::new()];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 640);
         assert!(out[1].is_empty());
@@ -716,11 +783,11 @@ mod tests {
 
     #[test]
     fn make_resampler_fi_empty() {
-        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedIn::<f64>::new(44100, 48000, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 1024);
         let waves = vec![Vec::new(); 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert!(out[0].is_empty());
         assert!(out[1].is_empty());
@@ -729,21 +796,21 @@ mod tests {
     #[test]
     fn make_resampler_fio_unusualratio() {
         // asking for 1024 give the nearest which is 1029 -> 1120
-        let mut resampler = FftFixedInOut::<f64>::new(44100, 44110, 1024, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedInOut::<f64>::new(44100, 44110, 1024, 2).unwrap();
+        let frames = resampler.input_frames_next();
         let waves = vec![vec![0.0f64; frames]; 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 4411);
     }
 
     #[test]
     fn make_resampler_fo_unusualratio() {
-        let mut resampler = FftFixedOut::<f64>::new(44100, 44110, 1024, 2, 2);
-        let frames = resampler.nbr_frames_needed();
+        let mut resampler = FftFixedOut::<f64>::new(44100, 44110, 1024, 2, 2).unwrap();
+        let frames = resampler.input_frames_next();
         assert_eq!(frames, 4410);
         let waves = vec![vec![0.0f64; frames]; 2];
-        let out = resampler.process(&waves).unwrap();
+        let out = resampler.process(&waves, None).unwrap();
         assert_eq!(out.len(), 2);
         assert_eq!(out[0].len(), 1024);
     }
